use std::fmt::Debug;

use chumsky::chain::Chain;

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct TypeStack<T>(Vec<T>);

impl<T: Debug + PartialEq> TypeStack<T> {
    pub fn new() -> Self {
        TypeStack(vec![])
    }
    pub fn len(&self) -> usize {
        self.0.len()
    }
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
    pub fn pop(mut self) -> TypeStack<T> {
        let _ = self.0.pop();
        self
    }
    pub fn push(mut self, t: T) -> TypeStack<T> {
        self.0.push(t);
        self
    }

    /// tests if ts matches the top of the stack
    pub fn test_consume(mut self, ts: TypeStack<T>) -> Result<TypeStack<T>, String> {
        if ts.len() > self.len() {
            Err(format!("error during test: {ts:?} is bigger than {self:?}"))
        } else {
            for (index, t) in ts.0.iter().rev().enumerate() {
                let val = self.0.pop().unwrap();
                if val != *t {
                    return Err(format!("type mismatch between {self:?} and {ts:?}\n{t:?} doesn't match {val:?} at stack depth {index:?}"));
                }
            }
            Ok(self)
        }
    }

    /// tests if ts matches the top of the stack
    pub fn test(&self, ts: &TypeStack<T>) -> Result<(), String> {
        if ts.len() > self.len() {
            Err(format!("error during test: {ts:?} is bigger than {self:?}"))
        } else {
            for (index, t) in ts.0.iter().rev().enumerate() {
                let val = &self.0[self.0.len() - 1];
                if val != t {
                    return Err(format!("type mismatch between {self:?} and {ts:?}\n{t:?} doesn't match {val:?} at stack depth {index:?}"));
                }
            }
            Ok(())
        }
    }

    // pub fn test_many(mut self, mut tss: Vec<TypeStack<T>>) -> Result<TypeStack<T>, String> {
    //     if tss.is_empty() {
    //         return Ok(TypeStack::new());
    //     }
    //     for _i in 0..tss.len() {
    //         let ts = tss.pop().unwrap();
    //         self = self.test(ts)?;
    //         if self.is_ok() {
    //             return self;
    //         }
    //     }
    //     Err("did not match any types".to_string())
    // }

    pub fn append(&mut self, t: &mut Vec<T>) {
        self.0.append(t);
    }
}

impl<T: PartialEq + Debug> From<Vec<T>> for TypeStack<T> {
    fn from(value: Vec<T>) -> Self {
        TypeStack(value)
    }
}

pub trait Checkable<T: PartialEq + Debug> {
    fn check(&mut self, stack: TypeStack<T>) -> Result<TypeStack<T>, String>;
}
